package solutions.bitbadger.documents.java

import solutions.bitbadger.documents.Configuration
import solutions.bitbadger.documents.DocumentException
import solutions.bitbadger.documents.Parameter
import java.sql.Connection
import java.sql.ResultSet
import java.sql.SQLException
import java.util.*
import kotlin.jvm.Throws

object Custom {

    /**
     * Execute a query that returns a list of results
     *
     * @param query The query to retrieve the results
     * @param parameters Parameters to use for the query
     * @param clazz The class of the document to be returned
     * @param conn The connection over which the query should be executed
     * @param mapFunc The mapping function between the document and the domain item
     * @return A list of results for the given query
     * @throws DocumentException If parameters are invalid
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TDoc> list(
        query: String,
        parameters: Collection<Parameter<*>> = listOf(),
        clazz: Class<TDoc>,
        conn: Connection,
        mapFunc: (ResultSet, Class<TDoc>) -> TDoc
    ) = Parameters.apply(conn, query, parameters).use { Results.toCustomList(it, clazz, mapFunc) }

    /**
     * Execute a query that returns a list of results (creates connection)
     *
     * @param query The query to retrieve the results
     * @param parameters Parameters to use for the query
     * @param clazz The class of the document to be returned
     * @param mapFunc The mapping function between the document and the domain item
     * @return A list of results for the given query
     * @throws DocumentException If no connection string has been set, or if parameters are invalid
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TDoc> list(
        query: String,
        parameters: Collection<Parameter<*>> = listOf(),
        clazz: Class<TDoc>,
        mapFunc: (ResultSet, Class<TDoc>) -> TDoc
    ) = Configuration.dbConn().use { list(query, parameters, clazz, it, mapFunc) }

    /**
     * Execute a query that returns one or no results
     *
     * @param query The query to retrieve the results
     * @param parameters Parameters to use for the query
     * @param clazz The class of the document to be returned
     * @param conn The connection over which the query should be executed
     * @param mapFunc The mapping function between the document and the domain item
     * @return An `Optional` value, with the document if one matches the query
     * @throws DocumentException If parameters are invalid
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TDoc> single(
        query: String,
        parameters: Collection<Parameter<*>> = listOf(),
        clazz: Class<TDoc>,
        conn: Connection,
        mapFunc: (ResultSet, Class<TDoc>) -> TDoc
    ) = Optional.ofNullable(list("$query LIMIT 1", parameters, clazz, conn, mapFunc).singleOrNull())

    /**
     * Execute a query that returns one or no results
     *
     * @param query The query to retrieve the results
     * @param parameters Parameters to use for the query
     * @param clazz The class of the document to be returned
     * @param mapFunc The mapping function between the document and the domain item
     * @return The document if one matches the query, `null` otherwise
     * @throws DocumentException If no connection string has been set, or if parameters are invalid
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TDoc> single(
        query: String,
        parameters: Collection<Parameter<*>> = listOf(),
        clazz: Class<TDoc>,
        mapFunc: (ResultSet, Class<TDoc>) -> TDoc
    ) = Configuration.dbConn().use { single(query, parameters, clazz, it, mapFunc) }

    /**
     * Execute a query that returns no results
     *
     * @param query The query to retrieve the results
     * @param conn The connection over which the query should be executed
     * @param parameters Parameters to use for the query
     * @throws DocumentException If parameters are invalid or if the query fails
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun nonQuery(query: String, parameters: Collection<Parameter<*>> = listOf(), conn: Connection) {
        try {
            Parameters.apply(conn, query, parameters).use { it.executeUpdate() }
        } catch (ex: SQLException) {
            throw DocumentException("Unable to execute non-query: ${ex.message}", ex)
        }
    }

    /**
     * Execute a query that returns no results
     *
     * @param query The query to retrieve the results
     * @param parameters Parameters to use for the query
     * @throws DocumentException If no connection string has been set, if parameters are invalid, or if the query fails
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun nonQuery(query: String, parameters: Collection<Parameter<*>> = listOf()) =
        Configuration.dbConn().use { nonQuery(query, parameters, it) }

    /**
     * Execute a query that returns a scalar result
     *
     * @param query The query to retrieve the result
     * @param parameters Parameters to use for the query
     * @param conn The connection over which the query should be executed
     * @param mapFunc The mapping function between the document and the domain item
     * @return The scalar value from the query
     * @throws DocumentException If parameters are invalid or if the query fails
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <T : Any> scalar(
        query: String,
        parameters: Collection<Parameter<*>> = listOf(),
        clazz: Class<T>,
        conn: Connection,
        mapFunc: (ResultSet, Class<T>) -> T
    ) = Parameters.apply(conn, query, parameters).use { stmt ->
        try {
            stmt.executeQuery().use { rs ->
                rs.next()
                mapFunc(rs, clazz)
            }
        } catch (ex: SQLException) {
            throw DocumentException("Unable to retrieve scalar value: ${ex.message}", ex)
        }
    }

    /**
     * Execute a query that returns a scalar result
     *
     * @param query The query to retrieve the result
     * @param parameters Parameters to use for the query
     * @param mapFunc The mapping function between the document and the domain item
     * @return The scalar value from the query
     * @throws DocumentException If no connection string has been set, or if parameters are invalid
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <T : Any> scalar(
        query: String,
        parameters: Collection<Parameter<*>> = listOf(),
        clazz: Class<T>,
        mapFunc: (ResultSet, Class<T>) -> T
    ) = Configuration.dbConn().use { scalar(query, parameters, clazz, it, mapFunc) }
}
