package solutions.bitbadger.documents.scala

import solutions.bitbadger.documents.{Configuration, Parameter}

import java.sql.{Connection, ResultSet}
import scala.reflect.ClassTag
import scala.util.Using

object Custom:

  /**
   * Execute a query that returns a list of results
   *
   * @param query      The query to retrieve the results
   * @param parameters Parameters to use for the query
   * @param conn       The connection over which the query should be executed
   * @param mapFunc    The mapping function between the document and the domain item
   * @return A list of results for the given query
   * @throws DocumentException If parameters are invalid
   */
  def list[Doc](query: String, parameters: Seq[Parameter[?]], conn: Connection,
                mapFunc: (ResultSet, ClassTag[Doc]) => Doc)(using tag: ClassTag[Doc]): List[Doc] =
    Using(Parameters.apply(conn, query, parameters)) { stmt => Results.toCustomList[Doc](stmt, mapFunc) }.get

  /**
   * Execute a query that returns a list of results
   *
   * @param query   The query to retrieve the results
   * @param conn    The connection over which the query should be executed
   * @param mapFunc The mapping function between the document and the domain item
   * @return A list of results for the given query
   * @throws DocumentException If parameters are invalid
   */
  def list[Doc](query: String, conn: Connection, mapFunc: (ResultSet, ClassTag[Doc]) => Doc)
               (using tag: ClassTag[Doc]): List[Doc] =
    list(query, List(), conn, mapFunc)

  /**
   * Execute a query that returns a list of results (creates connection)
   *
   * @param query      The query to retrieve the results
   * @param parameters Parameters to use for the query
   * @param mapFunc    The mapping function between the document and the domain item
   * @return A list of results for the given query
   * @throws DocumentException If parameters are invalid
   */
  def list[Doc](query: String, parameters: Seq[Parameter[?]], mapFunc: (ResultSet, ClassTag[Doc]) => Doc)
               (using tag: ClassTag[Doc]): List[Doc] =
    Using(Configuration.dbConn()) { conn => list[Doc](query, parameters, conn, mapFunc) }.get

  /**
   * Execute a query that returns a list of results (creates connection)
   *
   * @param query   The query to retrieve the results
   * @param mapFunc The mapping function between the document and the domain item
   * @return A list of results for the given query
   * @throws DocumentException If parameters are invalid
   */
  def list[Doc](query: String, mapFunc: (ResultSet, ClassTag[Doc]) => Doc)(using tag: ClassTag[Doc]): List[Doc] =
    list(query, List(), mapFunc)

  /**
   * Execute a query that returns one or no results
   *
   * @param query      The query to retrieve the results
   * @param parameters Parameters to use for the query
   * @param conn       The connection over which the query should be executed
   * @param mapFunc    The mapping function between the document and the domain item
   * @return An `Option` value, with the document if one matches the query
   * @throws DocumentException If parameters are invalid
   */
  def single[Doc](query: String, parameters: Seq[Parameter[?]], conn: Connection,
                  mapFunc: (ResultSet, ClassTag[Doc]) => Doc)(using tag: ClassTag[Doc]): Option[Doc] =
    list[Doc](s"$query LIMIT 1", parameters, conn, mapFunc).headOption

  /**
   * Execute a query that returns one or no results
   *
   * @param query   The query to retrieve the results
   * @param conn    The connection over which the query should be executed
   * @param mapFunc The mapping function between the document and the domain item
   * @return An `Option` value, with the document if one matches the query
   * @throws DocumentException If parameters are invalid
   */
  def single[Doc](query: String, conn: Connection, mapFunc: (ResultSet, ClassTag[Doc]) => Doc)
                 (using tag: ClassTag[Doc]): Option[Doc] =
    list[Doc](s"$query LIMIT 1", List(), conn, mapFunc).headOption

  /**
   * Execute a query that returns one or no results (creates connection)
   *
   * @param query      The query to retrieve the results
   * @param parameters Parameters to use for the query
   * @param mapFunc    The mapping function between the document and the domain item
   * @return An `Option` value, with the document if one matches the query
   * @throws DocumentException If parameters are invalid
   */
  def single[Doc](query: String, parameters: Seq[Parameter[?]], mapFunc: (ResultSet, ClassTag[Doc]) => Doc)
                 (using tag: ClassTag[Doc]): Option[Doc] =
    Using(Configuration.dbConn()) { conn => single[Doc](query, parameters, conn, mapFunc) }.get

  /**
   * Execute a query that returns one or no results (creates connection)
   *
   * @param query   The query to retrieve the results
   * @param mapFunc The mapping function between the document and the domain item
   * @return An `Option` value, with the document if one matches the query
   * @throws DocumentException If parameters are invalid
   */
  def single[Doc](query: String, mapFunc: (ResultSet, ClassTag[Doc]) => Doc)(using tag: ClassTag[Doc]): Option[Doc] =
    single[Doc](query, List(), mapFunc)

  /**
   * Execute a query that returns no results
   *
   * @param query      The query to retrieve the results
   * @param parameters Parameters to use for the query
   * @param conn       The connection over which the query should be executed
   * @throws DocumentException If parameters are invalid
   */
  def nonQuery(query: String, parameters: Seq[Parameter[?]], conn: Connection): Unit =
    Using(Parameters.apply(conn, query, parameters)) { stmt => stmt.executeUpdate() }

  /**
   * Execute a query that returns no results
   *
   * @param query The query to retrieve the results
   * @param conn  The connection over which the query should be executed
   * @throws DocumentException If parameters are invalid
   */
  def nonQuery(query: String, conn: Connection): Unit =
    nonQuery(query, List(), conn)

  /**
   * Execute a query that returns no results (creates connection)
   *
   * @param query      The query to retrieve the results
   * @param parameters Parameters to use for the query
   * @throws DocumentException If parameters are invalid
   */
  def nonQuery(query: String, parameters: Seq[Parameter[?]]): Unit =
    Using(Configuration.dbConn()) { conn => nonQuery(query, parameters, conn) }

  /**
   * Execute a query that returns no results (creates connection)
   *
   * @param query The query to retrieve the results
   * @throws DocumentException If parameters are invalid
   */
  def nonQuery(query: String): Unit =
    nonQuery(query, List())

  /**
   * Execute a query that returns a scalar result
   *
   * @param query      The query to retrieve the result
   * @param parameters Parameters to use for the query
   * @param conn       The connection over which the query should be executed
   * @param mapFunc    The mapping function between the document and the domain item
   * @return The scalar value from the query
   * @throws DocumentException If parameters are invalid
   */
  def scalar[A](query: String, parameters: Seq[Parameter[?]], conn: Connection, mapFunc: (ResultSet, ClassTag[A]) => A)
               (using tag: ClassTag[A]): A =
    Using(Parameters.apply(conn, query, parameters)) { stmt =>
      Using(stmt.executeQuery()) { rs =>
        rs.next()
        mapFunc(rs, tag)
      }.get
    }.get

  /**
   * Execute a query that returns a scalar result
   *
   * @param query   The query to retrieve the result
   * @param conn    The connection over which the query should be executed
   * @param mapFunc The mapping function between the document and the domain item
   * @return The scalar value from the query
   * @throws DocumentException If parameters are invalid
   */
  def scalar[A](query: String, conn: Connection, mapFunc: (ResultSet, ClassTag[A]) => A)(using tag: ClassTag[A]): A =
    scalar[A](query, List(), conn, mapFunc)

  /**
   * Execute a query that returns a scalar result (creates connection)
   *
   * @param query      The query to retrieve the result
   * @param parameters Parameters to use for the query
   * @param mapFunc    The mapping function between the document and the domain item
   * @return The scalar value from the query
   * @throws DocumentException If parameters are invalid
   */
  def scalar[A](query: String, parameters: Seq[Parameter[?]], mapFunc: (ResultSet, ClassTag[A]) => A)
               (using tag: ClassTag[A]): A =
    Using(Configuration.dbConn()) { conn => scalar[A](query, parameters, conn, mapFunc) }.get

  /**
   * Execute a query that returns a scalar result (creates connection)
   *
   * @param query   The query to retrieve the result
   * @param mapFunc The mapping function between the document and the domain item
   * @return The scalar value from the query
   * @throws DocumentException If parameters are invalid
   */
  def scalar[A](query: String, mapFunc: (ResultSet, ClassTag[A]) => A)(using tag: ClassTag[A]): A =
    scalar[A](query, List(), mapFunc)
